@@ -14,6 +14,8 @@
 # You should have received a copy of the GNU General Public License along with this
 # program.  If not, see <http://www.gnu.org/licenses/>.
 
+import ctypes
+import functools
 import glob
 import hashlib
 import importlib
@@ -21,12 +23,15 @@ import os
 import shutil
 import subprocess
 import sys
-import tempfile
 import threading
 import time
 import warnings
 from contextlib import contextmanager
-from functools import partial
+from functools import lru_cache, partial
+
+import numba  # noqa: F401
+import numpy as np  # noqa: F401
+from numba import float64, njit  # noqa: F401
 
 from .build_gsl import install_gsl_if_needed
 from .build_lsoda import install_lsoda_if_needed
@@ -44,6 +49,13 @@ from .utils import (
 HERE = os.path.dirname(os.path.abspath(__file__))
 
 
+ctypes.pythonapi.PyCapsule_GetPointer.restype = ctypes.c_void_p
+ctypes.pythonapi.PyCapsule_GetPointer.argtypes = [ctypes.py_object, ctypes.c_char_p]
+
+ctypes.pythonapi.PyCapsule_GetName.restype = ctypes.c_char_p
+ctypes.pythonapi.PyCapsule_GetName.argtypes = [ctypes.py_object]
+
+
 if os.environ.get("TOX_PACKAGE") is not None:
     print()
     print("*" * 80)
@@ -57,20 +69,26 @@ if os.environ.get("TOX_PACKAGE") is not None:
 
 
 def compile_if_needed_and_load(
-    module_wrapper, root_folder, lsoda_root_folder, gsl_root, compilation_flags
+    module_wrapper,
+    root_folder,
+    lsoda_root_folder,
+    gsl_root,
+    compilation_flags,
+    cache_folder=None,
 ):
-    cache_folder = base_cache_folder()
+    if cache_folder is None:
+        cache_folder = base_cache_folder()
 
     if gsl_root is None:
         gsl_root = os.path.join(cache_folder, "gsl", get_platform())
     if not os.path.exists(gsl_root) or not os.listdir(gsl_root):
-        download_folder = tempfile.mkdtemp()
+        download_folder = os.path.join(cache_folder, "gsl")
         install_gsl_if_needed(download_folder, gsl_root)
 
     lsoda_object_file = install_lsoda_if_needed(lsoda_root_folder)
     lsoda_fast_static_lib = install_lsoda_fast_if_needed(lsoda_root_folder)
 
-    libf2c_path = compile_libf2c()
+    libf2c_path = compile_libf2c(cache_folder)
 
     bf = build_folder(module_wrapper, compilation_flags, root_folder)
 
@@ -88,8 +106,8 @@ def compile_if_needed_and_load(
     return module
 
 
-def compile_libf2c():
-    folder = os.path.join(base_cache_folder(), "libf2c", get_platform())
+def compile_libf2c(cache_folder):
+    folder = os.path.join(cache_folder, "libf2c", get_platform())
     lib_path = os.path.join(folder, "libf2c.a")
     if os.path.exists(lib_path):
         return lib_path
@@ -132,7 +150,10 @@ def generate_files_and_compile_if_needed(
 
     create_folder_if_not_exists(build_folder)
 
-    if not glob.glob(os.path.join(build_folder, "_wrapper_{}.pyx".format(wrapper_id))):
+    build = os.environ.get("REBUILD_WRAPPER") or not glob.glob(
+        os.path.join(build_folder, "_wrapper_{}.pyx".format(wrapper_id))
+    )
+    if build:
         print("create source files at", build_folder)
         create_source_files(
             build_folder,
@@ -149,6 +170,12 @@ def generate_files_and_compile_if_needed(
     compile_files(build_folder)
 
 
+@lru_cache()
+def warn_new_traces(unique_id):
+    if read_new_traces(unique_id):
+        warnings.warn("there are new permuations pending, you might want to recompile")
+
+
 class LoadedModule:
     """wraps compiled module to support pickling"""
 
@@ -156,9 +183,9 @@ class LoadedModule:
         self._folder = os.path.abspath(folder)
         self._wrapper_id = os.path.basename(folder)
         self._module_name = "_wrapper_" + self._wrapper_id
-        self._init()
+        self._load()
 
-    def _init(self, state=None):
+    def _load(self, state=None):
         if self._module_name in sys.modules:
             del sys.modules[self._module_name]
         try:
@@ -173,15 +200,22 @@ class LoadedModule:
         finally:
             sys.path.pop(0)
             os.chdir(before)
-        for name, unique_id in self._module.get_fast_ode_unique_ids().items():
-            if read_new_traces(unique_id):
-                warnings.warn(
-                    "there are new permuations pending, you might want to recompile"
-                )
-                break
+        for unique_id in self._module.get_fast_ode_unique_ids().values():
+            warn_new_traces(unique_id)
+
+        self._setup_functions()
+
+    def _setup_functions(self):
+        for name in self._module._api:
+            _fun = setup_function(name, self._module)
+            setattr(self._module, name, _fun)
+
+        for name, f in self._module.__dict__.items():
+            if name.startswith("solve_fast_"):
+                setattr(self._module, name, functools.partial(f, self))
 
     def reload(self):
-        self._init()
+        self._load()
 
     def __getattr__(self, name):
         return getattr(self._module, name)
@@ -192,7 +226,7 @@ class LoadedModule:
     def __setstate__(self, data):
         try:
             self._folder, self._module_name, state = data
-            self._init(state)
+            self._load(state)
         except Exception:
             import traceback
 
@@ -202,6 +236,56 @@ class LoadedModule:
         return self._module.__dir__()
 
 
+def setup_function(name, module):
+    res_spec, arg_types, args, vector_flags, numba_signature = module._api[name]
+
+    capsule = module.__pyx_capi__[f"_api_{name}"]
+
+    fun_ptr = ctypes.pythonapi.PyCapsule_GetPointer(
+        capsule, ctypes.pythonapi.PyCapsule_GetName(capsule)
+    )
+    res_type, res_shape = res_spec
+    fun = ctypes.CFUNCTYPE(
+        ctypes.c_int,
+        *(arg_types + [res_type]),
+    )(fun_ptr)
+
+    ctype_args = ", ".join(
+        [
+            f"{arg}" + (".ctypes" if is_vector else "")
+            for (is_vector, arg) in zip(vector_flags, args)
+        ]
+        + ["__result.ctypes"]
+    )
+    arg_decl = ", ".join(args)
+    numba_decorator = (
+        f"@numba.njit([{numba_signature}])"
+        if not vector_flags or any(vector_flags)
+        else f"@numba.vectorize([{numba_signature}])"
+    )
+    code = align(
+        f"""
+    |{numba_decorator}
+    |def {name}({arg_decl}):
+    |    __result = np.zeros({res_shape if res_shape is not None else (1,)})
+    |
+    |    retcode = fun({ctype_args})
+    |    if retcode == 1:
+    |        raise ArithmeticError("calling {name} did not succeed, "
+    |                 "use get_last_error() function for details.")
+    |    if retcode == 2:
+    |        print("calling {name} issued a warning, "
+    |              "use get_last_warning() function for details.")
+    |    return {"__result[0]" if res_shape is None else  "__result"}
+    """
+    )
+    lo = dict(locals())
+    gl = dict(globals())
+    gl["fun"] = fun
+    exec(code, gl, lo)
+    return lo[name]
+
+
 # for backwards compatibility:
 load_module = LoadedModule
 
@@ -301,11 +385,12 @@ def create_source_files(
 
 
 def compile_files(folder):
-    try:
-        LoadedModule(folder)
-        return
-    except ImportError:
-        pass
+    if not os.environ.get("REBUILD_WRAPPER"):
+        try:
+            LoadedModule(folder)
+            return
+        except ImportError:
+            pass
     current_folder = os.getcwd()
     try:
         os.chdir(folder)
