@@ -31,10 +31,6 @@ def expression_hash(e):
     return hashlib.md5(str(e.sort_key()).encode("utf-8")).digest()
 
 
-def is_symbol_sublist(li1, li2):
-    return " ".join(map(str, li1)) in " ".join(map(str, li2))
-
-
 class Alias:
     def __init__(self, name, *symbols):
         assert isinstance(name, str)
@@ -154,19 +150,35 @@ class _FunctionWrapperBase(WrapperBase):
             var[0] if isinstance(var, tuple) else var for var in self.variables
         ]
 
-        self.args = ", ".join(
-            "double {}".format(var.c_args_decl()) for var in self.in_variables
+        if 0:
+            self.args = ", ".join(
+                "double {}".format(var.c_args_decl()) for var in self.in_variables
+            )
+
+        if 0:
+            self.cython_args = ", ".join(
+                "{}".format(var.cython_decl()) for var in self.in_variables
+            )
+
+        self.c_args_decl = ", ".join(
+            ["{}".format(var.c_decl()) for var in self.in_variables]
+            + ["double * __result"]
         )
 
-        self.cython_args = ", ".join(
-            "{}".format(var.cython_decl()) for var in self.in_variables
+        self.ctypes_args = ", ".join(
+            "{}".format(var.ctypes_decl()) for var in self.in_variables
         )
 
-        self.ufunc_args = ", ".join(
-            "double[{}] {}".format(":, :" if var.is_vector else ":", var._name())
-            for var in self.in_variables
+        self.numba_type_decl = ", ".join(
+            "{}".format(var.numba_type_decl()) for var in self.in_variables
         )
 
+        if 1:
+            self.ufunc_args = ", ".join(
+                "double[{}] {}".format(":, :" if var.is_vector else ":", var._name())
+                for var in self.in_variables
+            )
+
         common_expressions, simplified_expressions = sp.cse(
             self.expressions, ignore=list(bound_symbols(*self.expressions))
         )
@@ -209,12 +221,12 @@ class MatrixValuedFunctionWrapper(_FunctionWrapperBase):
     def setup_code_generation(self):
         super().setup_code_generation()
 
-        self._args = self.args
-
-        if self.args:
-            self.args = "{}, double _result[{}]".format(self.args, self.n * self.m)
-        else:
-            self.args = "double _result[{}]".format(self.n * self.m)
+        if 0:
+            self._args = self.args
+            if self.args:
+                self.args = "{}, double _result[{}]".format(self.args, self.n * self.m)
+            else:
+                self.args = "double _result[{}]".format(self.n * self.m)
 
     @concat_generator_results
     def c_code(self, header_file_path):
@@ -250,17 +262,14 @@ class MatrixValuedFunctionWrapper(_FunctionWrapperBase):
         )
 
         set_results = "\n|    ".join(
-            "_result[{}] = {};".format(i, expr)
+            "__result[{}] = {};".format(i, expr)
             for (i, expr) in enumerate(self.simplified_expressions)
             if expr != "0"
         )
 
-        sympy_expressions = "\n|    // ".join(str(e) for e in self.expressions)
-
         yield align(
             """
         |extern "C" void _{name}({args}) {{
-        |    // {sympy_expressions}
         |    {declare_arrays}
         |    {declare_scalars}
         |    {set_aliases}
@@ -269,8 +278,7 @@ class MatrixValuedFunctionWrapper(_FunctionWrapperBase):
         |}}
         """.format(
                 name=self.name,
-                args=self.args,
-                sympy_expressions=sympy_expressions,
+                args=self.c_args_decl,
                 declare_arrays=declare_arrays,
                 declare_scalars=declare_scalars,
                 set_aliases=set_aliases,
@@ -287,7 +295,7 @@ class MatrixValuedFunctionWrapper(_FunctionWrapperBase):
             """
         |extern "C" void _{name}({args});
         """
-        ).format(name=self.name, args=self.args)
+        ).format(name=self.name, args=self.c_args_decl)
 
     @concat_generator_results
     def cython_code(self, header_file_path):
@@ -309,7 +317,7 @@ class MatrixValuedFunctionWrapper(_FunctionWrapperBase):
         |cdef extern from "{header_file_path}":
         |    double _{name}({args})
         """.lstrip().format(
-                header_file_path=header_file_path, name=self.name, args=self.args
+                header_file_path=header_file_path, name=self.name, args=self.c_args_decl
             )
         )
 
@@ -321,6 +329,35 @@ class MatrixValuedFunctionWrapper(_FunctionWrapperBase):
         else:
             c_args = "&result[0]"
         signature = _signature(self.expressions)
+        yield align(
+            """
+        |cdef api int _api_{name}({args}):
+        |    cdef np.ndarray[np.double_t, ndim=1] result
+        |    cdef const char * _error_message
+        |    cdef const char * _warning_message
+        |
+        |    clear_last_error_message()
+        |    clear_last_warning_message()
+        |
+        |    result = np.zeros(({n} * {m},), dtype=np.double)
+        |    _{name}({c_args})
+
+        |    _error_message = c_get_last_error_message()
+        |    if _error_message[0] != 0:
+        |          return 1
+        |    _warning_message = c_get_last_warning_message()
+        |    if _warning_message[0] != 0:
+        |          return 2
+        |    return 0
+        |
+        """.lstrip().format(
+                name=self.name,
+                args=self.c_args_decl,
+                c_args=c_args,
+                m=self.m,
+                n=self.n,
+            )
+        )
         yield align(
             """
         |def {name}({args}):
@@ -334,10 +371,10 @@ class MatrixValuedFunctionWrapper(_FunctionWrapperBase):
         |    result = np.zeros(({n} * {m},), dtype=np.double)
         |    _{name}({c_args})
         |
-        |    _warning_message = get_last_warning_message()
+        |    _warning_message = c_get_last_warning_message()
         |    if _warning_message[0] != 0:
         |        warnings.warn(str(_warning_message, encoding="utf-8"))
-        |    _error_message = get_last_error_message()
+        |    _error_message = c_get_last_error_message()
         |    if _error_message[0] != 0:
         |          raise ArithmeticError(str(_error_message, encoding="utf-8") +
         |                " when evaluating {signature}"
@@ -376,49 +413,45 @@ class VectorValuedFunctionWrapper(MatrixValuedFunctionWrapper):
     def setup_code_generation(self):
         _FunctionWrapperBase.setup_code_generation(self)
 
-        self._args = self.args
-        if self.args:
-            self.args = "{}, double _result[{}]".format(self.args, self.n)
-        else:
-            self.args = "double _result[{}]".format(self.n)
-
     def cython_function_wrapper(self):
 
-        values = ", ".join(str(v.as_function_arg()) for v in self.in_variables)
-        if values:
-            c_args = values + ", &result[0]"
-        else:
-            c_args = "&result[0]"
-        signature = _signature(self.expressions)
+        args = [str(v.as_function_arg()) for v in self.in_variables]
+        values = ", ".join(args + ["__result"])
+
+        numba_signature = f"float64[:]({self.numba_type_decl})"
+        vector_flags = [v.is_vector for v in self.in_variables]
+
+        yield align(
+            f"""
+                |_api["{self.name}"] = ((ctypes.POINTER(ctypes.c_double), ({self.n},)),
+                |                       [{self.ctypes_args}],
+                |                       {args}, {vector_flags}, "{numba_signature}")
+                """
+        )
+
         yield align(
             """
-        |def {name}({args}):
-        |    cdef np.ndarray[np.double_t, ndim=1] result
+        |cdef api int _api_{name}({args}):
+        |    # cdef np.ndarray[np.double_t, ndim=1] result
         |    cdef const char * _error_message
         |    cdef const char * _warning_message
         |
         |    clear_last_error_message()
         |    clear_last_warning_message()
         |
-        |    result = np.zeros(({n},), dtype=np.double)
-        |    _{name}({c_args})
-        |
-        |    _warning_message = get_last_warning_message()
-        |    if _warning_message[0] != 0:
-        |        warnings.warn(str(_warning_message, encoding="utf-8"))
-        |    _error_message = get_last_error_message()
+        |    _{name}({values})
+
+        |    _error_message = c_get_last_error_message()
         |    if _error_message[0] != 0:
-        |          raise ArithmeticError(str(_error_message, encoding="utf-8") +
-        |                " when evaluating {signature}"
-        |                )
-        |
-        |    return result
+        |          return 1
+        |    _warning_message = c_get_last_warning_message()
+        |    if _warning_message[0] != 0:
+        |          return 2
+        |    return 0
         """.lstrip().format(
                 name=self.name,
-                args=self.cython_args,
-                c_args=c_args,
-                signature=signature,
-                n=self.n,
+                args=self.c_args_decl,
+                values=values,
             )
         )
 
@@ -465,18 +498,17 @@ class RealValuedFunctionWrapper(_FunctionWrapperBase):
         # created python wrappers:
         result = align(
             """
-        |extern "C" double _{name}({args}) {{
-        |    // {sympy_expression}
+        |extern "C" void _{name}({args}) {{
         |    {declare_arrays}
         |    {declare_scalars}
         |    {set_aliases}
         |    {set_common_vars}
-        |    return {c_expression};
+        |    *__result = {c_expression};
+        |
         |}}
         """.format(
                 name=self.name,
-                args=self.args,
-                sympy_expression=self.expressions[0],
+                args=self.c_args_decl,
                 declare_arrays=declare_arrays,
                 declare_scalars=declare_scalars,
                 set_aliases=set_aliases,
@@ -492,9 +524,9 @@ class RealValuedFunctionWrapper(_FunctionWrapperBase):
         # cython functions:
         yield align(
             """
-        |extern "C" double _{name}({args});
+        |extern "C" void _{name}({args});
         """
-        ).format(name=self.name, args=self.args)
+        ).format(name=self.name, args=self.c_args_decl)
 
     @concat_generator_results
     def cython_code(self, header_file_path):
@@ -507,6 +539,7 @@ class RealValuedFunctionWrapper(_FunctionWrapperBase):
             yield code
         for code in self.cython_function_wrapper():
             yield code
+
         for code in self.cython_ufunc_wrapper():
             yield code
 
@@ -516,13 +549,14 @@ class RealValuedFunctionWrapper(_FunctionWrapperBase):
         |cdef extern from "{header_file_path}":
         |    double _{name}({args})
         """.lstrip().format(
-                header_file_path=header_file_path, name=self.name, args=self.args
+                header_file_path=header_file_path, name=self.name, args=self.c_args_decl
             )
         )
 
     def cython_function_wrapper(self):
 
-        values = ", ".join(str(v.as_function_arg()) for v in self.in_variables)
+        args = [str(v.as_function_arg()) for v in self.in_variables]
+        values = ", ".join(args + ["__result"])
         _checks = []
 
         for alias in self.aliases:
@@ -531,38 +565,42 @@ class RealValuedFunctionWrapper(_FunctionWrapperBase):
                 "assert len({}) == {}, '{}'".format(alias.name, len(alias.symbols), msg)
             )
 
-        checks = "\n|    ".join(_checks)
+        numba_signature = f"float64({self.numba_type_decl})"
+
+        vector_flags = [v.is_vector for v in self.in_variables]
+
+        yield align(
+            f"""
+                |_api["{self.name}"] = ((ctypes.POINTER(ctypes.c_double), None),
+                |                       [{self.ctypes_args}],
+                |                       {args}, {vector_flags}, "{numba_signature}")
+                """
+        )
 
-        signature = _signature(self.expressions[0])
         yield align(
             """
-        |def {name}({args}):
-        |    cdef double result
+        |cdef api int _api_{name}({args}):
+        |
         |    cdef const char * _error_message
         |    cdef const char * _warning_message
         |
         |    clear_last_error_message()
         |    clear_last_warning_message()
         |
-        |    {checks}
-        |    result = _{name}({values})
+        |    _{name}({values})
         |
-        |    _warning_message = get_last_warning_message()
-        |    if _warning_message[0] != 0:
-        |        warnings.warn(str(_warning_message, encoding="utf-8"))
-        |    _error_message = get_last_error_message()
+        |    _error_message = c_get_last_error_message()
         |    if _error_message[0] != 0:
-        |          raise ArithmeticError(str(_error_message, encoding="utf-8") +
-        |                " when evaluating {signature}"
-        |                )
+        |          return 1
+        |    _warning_message = c_get_last_warning_message()
+        |    if _warning_message[0] != 0:
+        |          return 2
+        |    return 0
         |
-        |    return result
         """.lstrip().format(
                 name=self.name,
-                checks=checks,
-                args=self.cython_args,
+                args=self.c_args_decl,
                 values=values,
-                signature=signature,
             )
         )
 
@@ -577,25 +615,28 @@ class RealValuedFunctionWrapper(_FunctionWrapperBase):
             "|        raise ValueError('lengths must be one or agree.')"
         ).format(", ".join(f"{vi._name()}.shape[0]" for vi in self.in_variables))
 
-        out_size_computation = ("|    cdef size_t out_size = max({}, 0)").format(
+        out_size_computation = ("|    cdef size_t __out_size = max({}, 0)").format(
             ", ".join(f"{vi._name()}.shape[0]" for vi in self.in_variables)
         )
 
         values_decl = "\n|    ".join(
-            "cdef double{} _v{}".format(
+            "cdef double{} __v{}".format(
                 "[:]" if self.in_variables[i].is_vector else "", i
             )
             for i in range(out_dim)
         )
 
         set_values = "\n|        ".join(
-            f"_v{j} = {arg._name()}[0] if {arg}.shape[0] == 1 else {arg._name()}[i]"
+            f"__v{j} = {arg._name()}[0] if {arg}.shape[0] == 1 else {arg._name()}[__i]"
             for (j, arg) in enumerate(self.in_variables)
         )
 
         values = ", ".join(
-            f"&_v{j}[0]" if var.is_vector else f"_v{j}"
-            for j, var in enumerate(self.in_variables)
+            [
+                f"&__v{j}[0]" if var.is_vector else f"__v{j}"
+                for j, var in enumerate(self.in_variables)
+            ]
+            + ["&__result_view[__i]"]
         )
 
         signature = _signature(self.expressions[0])
@@ -605,32 +646,31 @@ class RealValuedFunctionWrapper(_FunctionWrapperBase):
         |@cython.boundscheck(False)
         |@cython.wraparound(False)
         |def {name}_ufunc({ufunc_args}):
-        |    cdef np.ndarray[np.double_t, ndim=1, cast=True] result
-        |    cdef const char * _error_message
-        |    cdef const char * _warning_message
-        |    cdef size_t i
+        |    cdef np.ndarray[np.double_t, ndim=1, cast=True] __result
+        |    cdef str __error_message, __warning_message
+        |    cdef size_t __i
         {sizes_check}
         {out_size_computation}
         |    {values_decl}
-        |    result = np.empty((out_size,), dtype=np.double)
-        |    cdef double[:] result_view = result
+        |    __result = np.empty((__out_size,), dtype=np.double)
+        |    cdef double[:] __result_view = __result
         |    clear_last_warning_message()
         |    clear_last_error_message()
         |
-        |    for i in range(out_size):
+        |    for __i in range(__out_size):
         |        {set_values}
-        |        result_view[i] = _{name}({values})
+        |        _{name}({values})
         |
-        |    _warning_message = get_last_warning_message()
-        |    if _warning_message[0] != 0:
-        |        warnings.warn(str(_warning_message, encoding="utf-8"))
+        |    __warning_message = get_last_warning_message()
+        |    if __warning_message:
+        |        warnings.warn(str(__warning_message, encoding="utf-8"))
         |
-        |    _error_message = get_last_error_message()
-        |    if _error_message[0] != 0:
-        |          raise ArithmeticError(str(_error_message, encoding="utf-8") +
+        |    __error_message = get_last_error_message()
+        |    if __error_message:
+        |          raise ArithmeticError(str(__error_message, encoding="utf-8") +
         |                " when evaluating {signature}"
         |                )
-        |    return result
+        |    return __result
         """.lstrip().format(
                 name=self.name,
                 ufunc_args=self.ufunc_args,
