@@ -546,13 +546,14 @@ def _general_cython_code(unique_id):
            |# distutils: language = c++
            |
            |cimport libc.stdio as stdio
-           |from libc.stdio cimport printf
+           |from libc.stdio cimport printf, fflush, stdout
            |from libc.stdlib cimport malloc
            |from posix.time cimport timeval, gettimeofday
            |from libcpp.vector cimport vector
            |from libcpp.set cimport set as cset
            |from libcpp.string cimport string
            |from libcpp.map cimport map
+           |from libcpp.unordered_map cimport unordered_map
            |from cython.operator cimport dereference as deref, preincrement as inc
            |
            |cimport cython
@@ -1102,7 +1103,8 @@ def _lsoda_fast_related_cython_code():
         |        double *t, double *tout, long int *itol, double *rtol,
         |        double *atol, long int *itask, long int *istate, long int *iopt,
         |        double *rwork, long int *lrw, long int *iwork, long int *liw,
-        |        U_fp jac, long int *jt, V_fp lu_solver, D_fp fnorm_computation) nogil
+        |        U_fp jac, long int *jt, V_fp lu_solver, D_fp fnorm_computation,
+        |        V_fp step_size_callback, long *handle) nogil
         |
         |    cdef struct ls0001_struct:
         |        double rowns[209], ccmax, el0, h, hmin, hmxi, hu, rc, tn, uround
@@ -1147,5 +1149,12 @@ def _lsoda_fast_related_cython_code():
         |
         |def set_sec_factor(double factor):
         |    _set_sec_factor(factor)
+        |
+        |ctypedef vector[double]* vec_double_ptr
+        |cdef unordered_map[long, vec_double_ptr] _hvalues = unordered_map[long, vec_double_ptr ]()
+        |
+        |cdef void _hvalue_callback(long *handle, double *h) nogil:
+        |    cdef vector[double]* hvalues_c = _hvalues.at(deref(handle))
+        |    deref(hvalues_c).push_back(deref(h))
         """
     )
